@@ -17,25 +17,34 @@
 
 package org.apache.commons.dbcp2;
 
+import static org.junit.Assert.*;
+
 import java.io.IOException;
+import java.lang.management.ManagementFactory;
 import java.sql.Connection;
 import java.sql.SQLException;
 import java.util.Arrays;
 import java.util.Properties;
+import java.util.Set;
 import java.util.concurrent.atomic.AtomicInteger;
 
+import javax.management.MBeanServer;
+import javax.management.ObjectName;
 import javax.sql.DataSource;
 
+import org.apache.commons.logging.LogFactory;
+import org.junit.After;
+import org.junit.Before;
+import org.junit.BeforeClass;
+import org.junit.Test;
+
 /**
  * TestSuite for BasicDataSource
  *
  * @author Dirk Verbeeck
- * @version $Revision: 1587109 $ $Date: 2014-04-13 15:07:42 -0700 (Sun, 13 Apr 2014) $
+ * @version $Id: TestBasicDataSource.java 1659452 2015-02-13 04:00:39Z psteitz $
  */
 public class TestBasicDataSource extends TestConnectionPool {
-    public TestBasicDataSource(String testName) {
-        super(testName);
-    }
 
     @Override
     protected Connection getConnection() throws Exception {
@@ -45,9 +54,14 @@ public class TestBasicDataSource extends TestConnectionPool {
     protected BasicDataSource ds = null;
     private static final String CATALOG = "test catalog";
 
-    @Override
+    @BeforeClass
+    public static void setUpClass() {
+        // register a custom logger which supports inspection of the log messages
+        LogFactory.getFactory().setAttribute("org.apache.commons.logging.Log", "org.apache.commons.dbcp2.StackMessageLog");
+    }
+
+    @Before
     public void setUp() throws Exception {
-        super.setUp();
         ds = createDataSource();
         ds.setDriverClassName("org.apache.commons.dbcp2.TesterDriver");
         ds.setUrl("jdbc:apache:commons:testdriver");
@@ -70,12 +84,14 @@ public class TestBasicDataSource extends TestConnectionPool {
     }
 
     @Override
+    @After
     public void tearDown() throws Exception {
         super.tearDown();
         ds.close();
         ds = null;
     }
 
+    @Test
     public void testClose() throws Exception {
         ds.setAccessToUnderlyingConnectionAllowed(true);
 
@@ -125,6 +141,7 @@ public class TestBasicDataSource extends TestConnectionPool {
 
     }
 
+    @Test
     public void testSetProperties() throws Exception {
         // normal
         ds.setConnectionProperties("name1=value1;name2=value2;name3=value3");
@@ -176,6 +193,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         }
     }
 
+    @Test
     public void testTransactionIsolationBehavior() throws Exception {
         Connection conn = getConnection();
         assertNotNull(conn);
@@ -195,12 +213,14 @@ public class TestBasicDataSource extends TestConnectionPool {
     }
 
     @Override
+    @Test
     public void testPooling() throws Exception {
         // this also needs access to the underlying connection
         ds.setAccessToUnderlyingConnectionAllowed(true);
         super.testPooling();
     }
 
+    @Test
     public void testNoAccessToUnderlyingConnectionAllowed() throws Exception {
         // default: false
         assertEquals(false, ds.isAccessToUnderlyingConnectionAllowed());
@@ -213,6 +233,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertNull(dconn);
     }
 
+    @Test
     public void testAccessToUnderlyingConnectionAllowed() throws Exception {
         ds.setAccessToUnderlyingConnectionAllowed(true);
         assertEquals(true, ds.isAccessToUnderlyingConnectionAllowed());
@@ -227,6 +248,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertTrue(dconn instanceof TesterConnection);
     }
 
+    @Test
     public void testEmptyValidationQuery() throws Exception {
         assertNotNull(ds.getValidationQuery());
 
@@ -237,6 +259,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertNull(ds.getValidationQuery());
     }
 
+    @Test
     public void testInvalidValidationQuery() {
         ds.setValidationQuery("invalid");
         try (Connection c = ds.getConnection()) {
@@ -248,6 +271,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         }
     }
 
+    @Test
     public void testValidationQueryTimoutFail() {
         ds.setTestOnBorrow(true);
         ds.setValidationQueryTimeout(3); // Too fast for TesterStatement
@@ -260,6 +284,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         }
     }
 
+    @Test
     public void testValidationQueryTimeoutZero() throws Exception {
         ds.setTestOnBorrow(true);
         ds.setTestOnReturn(true);
@@ -268,6 +293,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         con.close();
     }
 
+    @Test
     public void testValidationQueryTimeoutNegative() throws Exception {
         ds.setTestOnBorrow(true);
         ds.setTestOnReturn(true);
@@ -276,6 +302,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         con.close();
     }
 
+    @Test
     public void testValidationQueryTimeoutSucceed() throws Exception {
         ds.setTestOnBorrow(true);
         ds.setTestOnReturn(true);
@@ -284,6 +311,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         con.close();
     }
 
+    @Test
     public void testEmptyInitConnectionSql() throws Exception {
         ds.setConnectionInitSqls(Arrays.asList(new String[]{"", "   "}));
         assertNotNull(ds.getConnectionInitSqls());
@@ -294,6 +322,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertEquals(0, ds.getConnectionInitSqls().size());
     }
 
+    @Test
     public void testInvalidConnectionInitSql() {
         try {
             ds.setConnectionInitSqls(Arrays.asList(new String[]{"SELECT 1","invalid"}));
@@ -307,6 +336,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         }
     }
 
+    @Test
     public void testSetValidationTestProperties() {
         // defaults
         assertEquals(true, ds.getTestOnBorrow());
@@ -328,6 +358,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertEquals(false, ds.getTestWhileIdle());
     }
 
+    @Test
     public void testDefaultCatalog() throws Exception {
         Connection[] c = new Connection[getMaxTotal()];
         for (int i = 0; i < c.length; i++) {
@@ -352,6 +383,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         }
     }
 
+    @Test
     public void testSetAutoCommitTrueOnClose() throws Exception {
         ds.setAccessToUnderlyingConnectionAllowed(true);
         ds.setDefaultAutoCommit(Boolean.FALSE);
@@ -369,6 +401,7 @@ public class TestBasicDataSource extends TestConnectionPool {
         assertEquals(true, dconn.getAutoCommit());
     }
 
+    @Test
     public void testInitialSize() throws Exception {
         ds.setMaxTotal(20);
         ds.setMaxIdle(20);
@@ -384,13 +417,14 @@ public class TestBasicDataSource extends TestConnectionPool {
 
     // Bugzilla Bug 28251:  Returning dead database connections to BasicDataSource
     // isClosed() failure blocks returning a connection to the pool
+    @Test
     public void testIsClosedFailure() throws SQLException {
         ds.setAccessToUnderlyingConnectionAllowed(true);
         Connection conn = ds.getConnection();
         assertNotNull(conn);
         assertEquals(1, ds.getNumActive());
 
-        // set an IO failure causing the isClosed mathod to fail
+        // set an IO failure causing the isClosed method to fail
         TesterConnection tconn = (TesterConnection) ((DelegatingConnection<?>)conn).getInnermostDelegate();
         tconn.setFailure(new IOException("network error"));
 
@@ -402,12 +436,68 @@ public class TestBasicDataSource extends TestConnectionPool {
 
         assertEquals(0, ds.getNumActive());
     }
+    
+    /**
+     * Verifies correct handling of exceptions generated by the underlying pool as it closes
+     * connections in response to BDS#close. Exceptions have to be either swallowed by the
+     * underlying pool and logged, or propagated and wrapped.
+     */
+    @Test
+    public void testPoolCloseCheckedException() throws Exception {
+        ds.setAccessToUnderlyingConnectionAllowed(true);  // Allow dirty tricks
+        
+        // Get an idle connection into the pool
+        Connection conn = ds.getConnection();
+        TesterConnection tc = (TesterConnection) ((DelegatingConnection<?>) conn).getInnermostDelegate();
+        conn.close();
+        
+        // After returning the connection to the pool, bork it.
+        // Don't try this at home - bad violation of pool contract!
+        tc.setFailure(new SQLException("bang"));
+        
+        // Now close Datasource, which will cause tc to be closed, triggering SQLE
+        // Pool 2.x swallows and logs exceptions on pool close.  Below verifies that
+        // Either exceptions get logged or wrapped appropriately.
+        try {
+            StackMessageLog.lock();
+            StackMessageLog.clear();
+            ds.close();  
+            // Exception must have been swallowed by the pool - verify it is logged
+            assertTrue(StackMessageLog.popMessage().indexOf("bang") > 0);
+        } catch (SQLException ex) {
+            assertTrue(ex.getMessage().indexOf("Cannot close") > 0);
+            assertTrue(ex.getCause().getMessage().indexOf("bang") > 0);
+        } finally {
+            StackMessageLog.unLock();
+        }     
+    }
+    
+    @Test
+    public void testPoolCloseRTE() throws Exception {
+        // RTE version of testPoolCloseCheckedException - see comments there.
+        ds.setAccessToUnderlyingConnectionAllowed(true);
+        Connection conn = ds.getConnection();
+        TesterConnection tc = (TesterConnection) ((DelegatingConnection<?>) conn).getInnermostDelegate();
+        conn.close();
+        tc.setFailure(new IllegalStateException("boom"));
+        try {
+            StackMessageLog.lock();
+            StackMessageLog.clear();
+            ds.close();  
+            assertTrue(StackMessageLog.popMessage().indexOf("boom") > 0);
+        } catch (IllegalStateException ex) {
+            assertTrue(ex.getMessage().indexOf("boom") > 0); // RTE is not wrapped by BDS#close
+        } finally {
+            StackMessageLog.unLock();
+        }
+    }
 
     /**
      * Bugzilla Bug 29054:
      * The BasicDataSource.setTestOnReturn(boolean) is not carried through to
      * the GenericObjectPool variable _testOnReturn.
      */
+    @Test
     public void testPropertyTestOnReturn() throws Exception {
         ds.setValidationQuery("select 1 from dual");
         ds.setTestOnBorrow(false);
@@ -427,6 +517,7 @@ public class TestBasicDataSource extends TestConnectionPool {
      * The DaffodilDB driver throws an SQLException if
      * trying to commit or rollback a readOnly connection.
      */
+    @Test
     public void testRollbackReadOnly() throws Exception {
         ds.setDefaultReadOnly(Boolean.TRUE);
         ds.setDefaultAutoCommit(Boolean.FALSE);
@@ -441,6 +532,7 @@ public class TestBasicDataSource extends TestConnectionPool {
      * MaxTotal == 0 should throw SQLException on getConnection.
      * Results from Bug 29863 in commons-pool.
      */
+    @Test
     public void testMaxTotalZero() throws Exception {
         ds.setMaxTotal(0);
 
@@ -453,10 +545,26 @@ public class TestBasicDataSource extends TestConnectionPool {
             // test OK
         }
     }
+    
+    @Test
+    public void testInvalidateConnection() throws Exception {
+    	ds.setMaxTotal(2);
+    	Connection conn1 = ds.getConnection();
+    	Connection conn2 = ds.getConnection();
+    	ds.invalidateConnection(conn1);
+    	assertTrue(conn1.isClosed());
+    	assertEquals(1, ds.getNumActive());
+    	assertEquals(0, ds.getNumIdle());
+    	Connection conn3 = ds.getConnection();
+    	conn2.close();
+    	conn3.close();
+    }
+    
     /**
      * JIRA DBCP-93: If an SQLException occurs after the GenericObjectPool is
      * initialized in createDataSource, the evictor task is not cleaned up.
      */
+    @Test
     public void testCreateDataSourceCleanupThreads() throws Exception {
         ds.close();
         ds = null;
@@ -490,6 +598,7 @@ public class TestBasicDataSource extends TestConnectionPool {
      * JIRA DBCP-333: Check that a custom class loader is used.
      * @throws Exception
      */
+    @Test
     public void testDriverClassLoader() throws Exception {
         getConnection();
         ClassLoader cl = ds.getDriverClassLoader();
@@ -503,6 +612,7 @@ public class TestBasicDataSource extends TestConnectionPool {
      * Verify that when errors occur during BasicDataSource initialization, GenericObjectPool
      * Evictors are cleaned up.
      */
+    @Test
     public void testCreateDataSourceCleanupEvictor() throws Exception {
         ds.close();
         ds = null;
@@ -543,6 +653,89 @@ public class TestBasicDataSource extends TestConnectionPool {
         // make sure cleanup is complete
         assertNull(ds.getConnectionPool());
     }
+    
+    @Test
+    public void testMaxConnLifetimeExceeded() throws Exception {
+        try {
+            StackMessageLog.lock();
+            ds.setMaxConnLifetimeMillis(100);
+            Connection conn = ds.getConnection();
+            assertEquals(1, ds.getNumActive());
+            Thread.sleep(500);
+            conn.close();
+            assertEquals(0, ds.getNumIdle());
+            String message = StackMessageLog.popMessage();
+            assertTrue(message.indexOf("exceeds the maximum permitted value") > 0);
+        } finally {
+            StackMessageLog.clear();
+            StackMessageLog.unLock();
+        }
+    }
+    
+    @Test
+    public void testMaxConnLifetimeExceededMutedLog() throws Exception {
+        try {
+            StackMessageLog.lock();
+            StackMessageLog.clear();
+            ds.setMaxConnLifetimeMillis(100);
+            ds.setLogExpiredConnections(false);
+            Connection conn = ds.getConnection();
+            assertEquals(1, ds.getNumActive());
+            Thread.sleep(500);
+            conn.close();
+            assertEquals(0, ds.getNumIdle());
+            assertTrue(StackMessageLog.isEmpty());
+        } finally {
+            StackMessageLog.clear();
+            StackMessageLog.unLock();
+        }
+    }
+    
+    @Test
+    public void testConcurrentInitBorrow() throws Exception {
+        ds.setDriverClassName("org.apache.commons.dbcp2.TesterConnectionDelayDriver");
+        ds.setUrl("jdbc:apache:commons:testerConnectionDelayDriver:50");
+        ds.setInitialSize(8);
+        
+        // Launch a request to trigger pool initialization
+        TestThread testThread = new TestThread(1,0);
+        Thread t = new Thread(testThread);
+        t.start();
+        
+        // Get another connection (should wait for pool init)
+        Thread.sleep(100); // Make sure t gets into init first
+        ds.getConnection();
+        
+        // Pool should have at least 6 idle connections now
+        // Use underlying pool getNumIdle to avoid waiting for ds lock
+        assertTrue(ds.getConnectionPool().getNumIdle() > 5);
+        
+        // Make sure t completes successfully
+        t.join();
+        assertFalse(testThread.failed());
+        
+        ds.close();
+    }
+    
+    /**
+     * Make sure setting jmxName to null suppresses JMX registration of connection and statement pools.
+     * JIRA: DBCP-434
+     */
+    @Test
+    public void testJmxDisabled() throws Exception {
+        MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
+        // Unregister leftovers from other tests (TODO: worry about concurrent test execution)
+        ObjectName commons = new ObjectName("org.apache.commons.*:*");
+        Set<ObjectName> results = mbs.queryNames(commons, null);
+        for (ObjectName result : results) {
+            mbs.unregisterMBean(result);
+        }
+        ds.setJmxName(null);  // Should disable JMX for both connection and statement pools
+        ds.setPoolPreparedStatements(true);
+        ds.getConnection();  // Trigger initialization
+        // Nothing should be registered
+        assertEquals(0, mbs.queryNames(commons, null).size());
+    }
 }
 
 /**
@@ -571,3 +764,28 @@ class TesterConnRequestCountDriver extends TesterDriver {
         connectionRequestCount.set(0);
     }
 }
+
+/**
+ * TesterDriver that adds latency to connection requests. Latency (in ms) is the
+ * last component of the URL.
+ */
+class TesterConnectionDelayDriver extends TesterDriver {
+    private static final String CONNECT_STRING = "jdbc:apache:commons:testerConnectionDelayDriver";
+
+    @Override
+    public Connection connect(String url, Properties info) throws SQLException {
+        String[] parsedUrl = url.split(":");
+        int delay = Integer.parseInt(parsedUrl[parsedUrl.length - 1]);
+        try {
+            Thread.sleep(delay);
+        } catch(InterruptedException ex) {
+            Thread.currentThread().interrupt();
+        }
+        return super.connect(url, info);
+    }
+
+    @Override
+    public boolean acceptsURL(String url) throws SQLException {
+        return url.startsWith(CONNECT_STRING);
+    }
+}
